// Fitness calculation functions
// Assigns internal fitness and returns value
double Problem::Fitness(Chromosone *template, Chromosone *test_chrom)
{
  double Fix_Fitness(long *);
  
  Chromosone *temp_chrom;    // temporary chromosone
  long *map;                 // fixed size mapping
  double fitness;            // fitness value
  
  // Make temp_chrom a composite of the template and test_chrom
  temp_chrom = template->Copy();
  temp_chrom->Splice(test_chrom);
  
  // make the mapping from the test_chromosone
  int count, maxlen;
  Gene *cur_gene;
  maxlen = temp_chrom->Get_Len();
  if ((map = (long *) malloc(maxlen * sizeof(long))) == (long *) NULL)
    {
      cout << "Problem::Fitness memory allocation failed for map\n";
      exit(0);
    }
  for(count=0,cur_gene = temp_chrom->first_gene;
      (count < maxlen);
      count++, cur_gene = cur_gene->Next())
      {
	if (cur_gene == (Gene *) NULL)
	  {
	    cout << "Problem::Fitness got an unexpected NULL gene\n";
	    exit(0);
	  }
	map[(int) cur_gene->Gene_id()] = cur_gene->Allele();
      }

  // Now we have a completed mapping array, so now let's use it
  // to get the fitness

  int node,side,target;
  int x1,x2,y1,y2,dx,dy,temp;
  double maxfit,newfit=0;
  edges = 0;

  for(node=0;(node < size);node++)
    for(side=0;(side < connect);side++)
      {
	target = edge[(node*connect)+side];  // edge[node][side]

	// if it's not a null edge, then compute change in fitness
	if (target != -1)
	  {
	    edges++;    // realize the mapping has an edge

	    // Modified Manhattan Distance
	    x1 = map[node]%mach_side;
	    x2 = map[target]%mach_side;
	    dx = abs(x1 - x2);
	    temp = ((x1>x2) ? x2+abs(mach_side-x1) :
		    x1+abs(mach_side-x2));
	    dx = ((dx>temp) ? temp : dx);

	    y1 = map[node]/mach_side;
	    y2 = map[target]/mach_side;
	    dy = abs(y1 - y2);
	    temp = ((y1>y2) ? y2 + abs(mach_side-y1) :
		    y1 + abs(mach_side-y2));
	    dy = ((dy>temp) ? temp : dy);
	    newfit += ((dx>dy) ? dy : dx) + abs(dx-dy);

	    //cout << "\n ("<<x1<<","<<y1<<") ("<<x2<<","<<y2<<") ";
	    //cout << "dx"<<dx<<" dy"<<dy<<" tot"<<newfit;
	  }
      }
  fitness = newfit/2.0;       // We count every edge twice, so rectify
  edges = edges/2;            // ditto

  // squash our fitness into the range [0,1]
  maxfit = (double) (edges * (mach_side/2));
  if (maxfit == ((double) edges))
    fitness = 1.0;
  else
    fitness = 1 - ((fitness-((double) edges))/(maxfit-((double) edges)));
  return(fitness);
}
